/*****************************
gemini.css

Example user CSS file for Planet Computers' Gemini PDA running LXQt.
Include from ~/.config/pygenda/pygenda.css or copy & adapt as desired.

******************************/

/* Remove transition animations - they make Gemini feel slow */
* {
	transition-duration:0s;
	transition-delay:0s;
}

/* Icon theme - see /usr/share/icons/ for installed themes */
image {
	-gtk-icon-theme:"Adwaita";
}

/* GUI size increases for Gemini's small screen */

menubar, button {
	font-size:15pt;
}

menubar > menuitem {
	margin-right:0.1em;
}

dialog {
	font-size:15pt;
}

/* Increasing font size in dialog distorts switches, so we tweak size... */
switch {
	border-radius:12pt;
}

switch slider {
	min-height:20pt;
	border-radius:10pt;
}

scrollbar.vertical, scrollbar.vertical slider, scrollbar button {
	min-width:14px;
}

scrollbar.horizontal, scrollbar.horizontal slider, scrollbar button {
	min-height:14px;
}

scrollbar trough {
	background-color:#f0f0f0;
}

scrollbar slider {
	background-color:#c0c0c0;
	border-radius:0;
}


/* Week View */

#weekview_titlearea {
	border-width:3px;
}

#weekview_labelmonth {
	font-size:21pt;
}

.zoom3 #weekview_labelmonth {
	font-size:22.5pt;
}

.zoom4 #weekview_labelmonth {
	font-size:24pt;
}

#weekview_labelweek {
	font-size:13pt;
}

.zoom3 #weekview_labelweek {
	font-size:14pt;
}

.zoom4 #weekview_labelweek {
	font-size:15pt;
}

.weekview_labelday {
	font-size:18pt;
	min-width:2.95em;
}

.zoom0 .weekview_daytext {
	font-size:13pt;
}

.zoom1 .weekview_daytext {
	font-size:14.5pt;
}

.zoom2 .weekview_daytext {
	font-size:16pt;
}

.zoom3 .weekview_daytext {
	font-size:17.5pt;
}

.zoom4 .weekview_daytext {
	font-size:19pt;
}

.zoom2 .weekview_marker {
	font-family:'DejaVu Sans Condensed';
}

.zoom3 .weekview_marker {
	font-family:'DejaVu Sans Condensed';
}

.zoom4 .weekview_marker {
	font-family:'DejaVu Sans Condensed';
}


/* Year View */

#yearview_labelyear {
	font-size:18pt;
	padding:2px;
}

#yearview_labeldate {
	padding:2px;
}

.yearview_daycell {
	font-size:9pt;
}

.zoom0 #yearview_datecontent, .zoom0 #yearview_labeldate {
	font-size:14pt;
}

.zoom1 #yearview_datecontent, .zoom1 #yearview_labeldate {
	font-size:15.5pt;
}

.zoom2 #yearview_datecontent, .zoom2 #yearview_labeldate {
	font-size:17pt;
}

.zoom3 #yearview_datecontent, .zoom3 #yearview_labeldate {
	font-size:18.5pt;
}

.zoom4 #yearview_datecontent, .zoom4 #yearview_labeldate {
	font-size:20pt;
}


/* Todo View */

.todoview_list {
	min-width:400px;
}

.zoom0 .todoview_title {
	font-size:14pt;
}

.zoom1 .todoview_title {
	font-size:15.5pt;
}

.zoom2 .todoview_title {
	font-size:17pt;
}

.zoom3 .todoview_title {
	font-size:18.5pt;
}

.zoom4 .todoview_title {
	font-size:20pt;
}

.zoom0 .todoview_items {
	font-size:13pt;
}

.zoom1 .todoview_items {
	font-size:14.5pt;
}

.zoom2 .todoview_items {
	font-size:16pt;
}

.zoom3 .todoview_items {
	font-size:17.5pt;
}

.zoom4 .todoview_items {
	font-size:19pt;
}

/* Dialogs */

entry, spinbutton, date_entry, time_entry, duration_entry {
	min-height:1.5em;
}

.longtextentry > textview {
	padding-top:.1ex;
	padding-bottom:.1ex;
}

/* Set height of Repeats tab, so expanding it doesn't expand Event dialog */
#grid-repeats {
	min-height:24.1ex;
}

/* Make buttons bigger (for easier touchscreen use) */
buttonbox > button {
	padding:5px;
}

buttonbox > button:focus {
	padding:4px;
}

checkbutton {
	min-height:32px;
}

check {
	min-height:24px;
	min-width:24px;
}

radio, check {
	border-style:solid;
	border-width:0px;
	padding:2px;
}

radio {
	border-radius:8px;
}

check {
	border-radius:4px;
}

radio:focus, check:focus {
	border-color:#398ee7;
	border-width:2px;
	padding:0px;
}

.llabel {
	min-height:34px;
	padding-right:0.3ex;
	padding-left:0.3ex;
}
