/*****************************
  pygenda.css
  Copyright (C) 2022-2024 Matthew Lewis

  This file is part of Pygenda.

  Pygenda is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, version 3.

  Pygenda is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with Pygenda. If not, see <https://www.gnu.org/licenses/>.
******************************/

/* Global */

scrollbar {
	border:0;
}

.softbutton {
	min-width:5em;
}

.view, #view_loading {
	min-width:600px;
	min-height:300px;
}

#view_loading spinner {
	min-width:32px;
}


/* Week View */

#view_week {
	background-color:white;
}

#weekview_titlearea {
	padding:3px;
	color:black;
	border-style:none none solid none;
	border-color:black;
	border-width:2px;
}

#weekview_labelmonth {
	font-family:'DejaVu Sans Condensed';
	font-size:12pt;
	font-weight:bold;
}

.zoom3 #weekview_labelmonth {
	font-size:13pt;
}

.zoom4 #weekview_labelmonth {
	font-size:14pt;
}

#weekview_labelweek {
	font-family:'DejaVu Sans';
	font-size:9pt;
}

.zoom3 #weekview_labelweek {
	font-size:9.5pt;
}

.zoom4 #weekview_labelweek {
	font-size:10pt;
}

#weekview_fold {
	min-width:12px;
	background-color:transparent;
	background-image:linear-gradient(90deg, rgba(255,255,255,0) 0%, rgba(136,136,136,0.8) 35%, rgba(0,0,0,1) 50%, rgba(136,136,136,0.8) 65%, rgba(255,255,255,0) 100%);
}

.weekview_day {
	border-style:none none solid none;
	border-color:#888;
	border-width:1px;
}

.weekview_day2, .weekview_day6 {
	border-style:none;
}

.weekview_labelday {
	background-color:rgba(190,190,190,0.8);
	color:black;
	font-family:'DejaVu Sans Condensed';
	font-size:10pt;
	font-weight:bold;
	min-width:4em;
}

.weekview_today .weekview_labelday {
	background-color:rgba(0,170,170,0.8);
}

.weekview_pastday .weekview_labelday {
	color:#555;
}

.weekview_daytext {
	padding:4px;
	color:black;
	font-family:'DejaVu Sans';
	font-weight:normal;
}

.zoom0 .weekview_daytext {
	font-size:8pt;
}

.zoom1 .weekview_daytext {
	font-size:9pt;
}

.zoom2 .weekview_daytext {
	font-size:10pt;
}

.zoom3 .weekview_daytext {
	font-size:11pt;
}

.zoom4 .weekview_daytext {
	font-size:12pt;
}

.weekview_marker {
	color:black;
	padding-left:1.5px;
	padding-right:1.5px;
	margin-right:0.2em;
	min-width:1ex;
}

.weekview_cursor {
	color:white;
	background-color:black;
	border-radius:0.3em;
}


/* Year View */

#view_year {
	color:black;
	background-color:white;
}

#yearview_labelyear {
	font-family:'DejaVu Sans';
	font-size:14pt;
	font-weight:bold;
}

.yearview_day_label, .yearview_month_label {
	font-weight:bold;
}

.yearview_month_label {
	padding-right:1px;
}

/* This sets a minimum height for cells in the Year View grid.
   May want to decrease this for small screens or increase for large. */
#yearview_daygrid {
	font-size:8pt;
}

.yearview_emptycell {
	background-color:rgba(210,210,210,0.4);
	border-right:solid #ddd 1px;
	border-bottom:solid #ddd 1px;
}

.yearview_daycell {
	font-family:'DejaVu Sans Condensed';
	font-size:7pt;
	background-color:rgba(255,255,255,0.7);
	border-right:solid black 1px;
	border-bottom:solid black 1px;
	padding:1px;
}

.yearview_leftofdaycell {
	border-right:solid black 1px;
}

.yearview_month_label.yearview_leftofdaycell {
	padding-right:0;
}

.yearview_abovedaycell {
	border-bottom:solid black 1px;
}

.yearview_day_sat, .yearview_day_sun {
	background-color:rgba(160,160,160,0.7);
}

.yearview_cursor {
	background-image:linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_today {
	background-color:#0aa;
}

.yearview_pastday {
	background-color:rgba(235,235,235,0.5);
}

.yearview_pastday.yearview_day_sat, .yearview_pastday.yearview_day_sun {
	background-color:rgba(160,160,160,0.5);
}

#yearview_labeldate {
	font-family:'DejaVu Sans Condensed';
	font-weight:bold;
	padding-right:0.8ex;
}

#yearview_datecontent {
	font-family:'DejaVu Sans';
}

.zoom0 #yearview_datecontent, .zoom0 #yearview_labeldate {
	font-size:9pt;
}

.zoom1 #yearview_datecontent, .zoom1 #yearview_labeldate {
	font-size:10pt;
}

.zoom2 #yearview_datecontent, .zoom2 #yearview_labeldate {
	font-size:11pt;
}

.zoom3 #yearview_datecontent, .zoom3 #yearview_labeldate {
	font-size:12pt;
}

.zoom4 #yearview_datecontent, .zoom4 #yearview_labeldate {
	font-size:13pt;
}

#yearview_datecontent_scroller {
	border:0;
}

.yearview_marker {
	padding-left:1.5px;
	padding-right:1.5px;
	margin-right:0.2em;
	min-width:1ex;
}

.yearview_entry_cursor {
	color:rgba(255,255,255,0.93);
	background-color:black;
	border-radius:0.3em;
}

.yearview_entry_single, .yearview_entry_repeated_day, .yearview_entry_repeated_hour, .yearview_entry_repeated_minute, .yearview_entry_repeated_second {
	background-image:url("disc.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_cursor, .yearview_entry_repeated_day.yearview_cursor, .yearview_entry_repeated_hour.yearview_cursor, .yearview_entry_repeated_minute.yearview_cursor, .yearview_entry_repeated_second.yearview_cursor {
	background-image:url("disc.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_repeated_month, .yearview_entry_repeated_week {
	background-image:url("loop.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_repeated_month.yearview_cursor, .yearview_entry_repeated_week.yearview_cursor {
	background-image:url("loop.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_repeated_year {
	background-image:url("star.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_repeated_year.yearview_cursor {
	background-image:url("star.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_todo {
	background-image:url("T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_todo.yearview_cursor {
	background-image:url("T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_repeated_month, .yearview_entry_single.yearview_entry_repeated_week, .yearview_entry_repeated_day.yearview_entry_repeated_month, .yearview_entry_repeated_day.yearview_entry_repeated_week, .yearview_entry_repeated_hour.yearview_entry_repeated_month, .yearview_entry_repeated_hour.yearview_entry_repeated_week, .yearview_entry_repeated_minute.yearview_entry_repeated_month, .yearview_entry_repeated_minute.yearview_entry_repeated_week, .yearview_entry_repeated_second.yearview_entry_repeated_month, .yearview_entry_repeated_second.yearview_entry_repeated_week {
	background-image:url("disc+loop.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_cursor, .yearview_entry_single.yearview_entry_repeated_week.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_cursor {
	background-image:url("disc+loop.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_repeated_year, .yearview_entry_repeated_day.yearview_entry_repeated_year, .yearview_entry_repeated_hour.yearview_entry_repeated_year, .yearview_entry_repeated_minute.yearview_entry_repeated_year, .yearview_entry_repeated_second.yearview_entry_repeated_year {
	background-image:url("disc+star.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_year.yearview_cursor {
	background-image:url("disc+star.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_todo, .yearview_entry_repeated_day.yearview_entry_todo, .yearview_entry_repeated_hour.yearview_entry_todo, .yearview_entry_repeated_minute.yearview_entry_todo, .yearview_entry_repeated_second.yearview_entry_todo {
	background-image:url("disc+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_todo.yearview_cursor {
	background-image:url("disc+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_repeated_month.yearview_entry_repeated_year, .yearview_entry_repeated_week.yearview_entry_repeated_year {
	background-image:url("loop+star.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_cursor {
	background-image:url("loop+star.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_repeated_month.yearview_entry_todo, .yearview_entry_repeated_week.yearview_entry_todo {
	background-image:url("loop+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_repeated_month.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_week.yearview_entry_todo.yearview_cursor {
	background-image:url("loop+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_repeated_year.yearview_entry_todo {
	background-image:url("star+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor {
	background-image:url("star+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_entry_repeated_year, .yearview_entry_single.yearview_entry_repeated_week.yearview_entry_repeated_year, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_entry_repeated_year, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_entry_repeated_year, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_entry_repeated_year, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_entry_repeated_year, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_entry_repeated_year, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_entry_repeated_year, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_entry_repeated_year, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_entry_repeated_year {
	background-image:url("disc+loop+star.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_single.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_cursor {
	background-image:url("disc+loop+star.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_entry_todo, .yearview_entry_single.yearview_entry_repeated_week.yearview_entry_todo, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_entry_todo, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_entry_todo, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_entry_todo, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_entry_todo, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_entry_todo, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_entry_todo, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_entry_todo, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_entry_todo {
	background-image:url("disc+loop+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_entry_todo.yearview_cursor, .yearview_entry_single.yearview_entry_repeated_week.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_entry_todo.yearview_cursor {
	background-image:url("disc+loop+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_day.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_hour.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_minute.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_second.yearview_entry_repeated_year.yearview_entry_todo {
	background-image:url("disc+star+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor {
	background-image:url("disc+star+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo {
	background-image:url("loop+star+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor {
	background-image:url("loop+star+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_single.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo {
	background-image:url("disc+loop+star+T.svg");
	background-size:100% auto;
	background-position:center;
	background-repeat:no-repeat;
}

.yearview_entry_single.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_single.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_day.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_hour.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_minute.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_month.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor, .yearview_entry_repeated_second.yearview_entry_repeated_week.yearview_entry_repeated_year.yearview_entry_todo.yearview_cursor {
	background-image:url("disc+loop+star+T.svg"), linear-gradient(90deg, #000 9%, transparent 9%, transparent 91%, #000 91%), linear-gradient(#000 8%, transparent 8%, transparent 92%, #000 92%);
}

/* Todo View */

.todoview_title {
	font-family:'DejaVu Sans';
	font-weight:bold;
}

.zoom0 .todoview_title {
	font-size:9pt;
}

.zoom1 .todoview_title {
	font-size:10pt;
}

.zoom2 .todoview_title {
	font-size:11pt;
}

.zoom3 .todoview_title {
	font-size:12pt;
}

.zoom4 .todoview_title {
	font-size:13pt;
}

.todoview_list {
	/* make lists half width of view-1, so two to-do lists fit nicely */
	min-width:299px;
}

.todoview_list:not(:first-child) {
	border-left:solid #888 1px;
	}

.todoview_items {
	padding:4px;
	font-family:'DejaVu Sans';
}

.todoview_item {
	padding-top:0.1em;
}

.zoom0 .todoview_items {
	font-size:8pt;
}

.zoom1 .todoview_items {
	font-size:9pt;
}

.zoom2 .todoview_items {
	font-size:10pt;
}

.zoom3 .todoview_items {
	font-size:11pt;
}

.zoom4 .todoview_items {
	font-size:12pt;
}

.todoview_marker {
	padding-left:1.5px;
	padding-right:1.5px;
	margin-right:0.2em;
	min-width:1ex;
}

.todoview_cursor {
	color:white;
	background-color:black;
	border-radius:0.3em;
}

.todo.needs-action, .needs-action > .todoview_itemtext {
	color:red;
	font-weight:bold;
}

.todo.completed, .completed > .todoview_itemtext {
	text-decoration:line-through;
}

.todo.in-process, .in-process > .todoview_itemtext {
	color: green;
	font-style:italic;
}

.todo.cancelled, .event.cancelled, .cancelled > .todoview_itemtext {
	color:gray;
	text-decoration:line-through #444;
}

/* Event statuses */

.tentative {
	color:gray;
}

.confirmed {
	font-weight:bold;
	/* font-style:italic; */
	/* text-decoration:underline; */
}


/* Dialogs */

entry, spinbutton, date_entry, time_entry, duration_entry {
	min-height:1.8em;
}

date_entry, time_entry, duration_entry {
	background-color:white;
	border:solid #888 1px;
	border-radius:3px;
}

date_entry > *, time_entry > *, duration_entry > * {
	background:transparent;
	border:none;
}

date_entry > label, time_entry > label, duration_entry > label {
	padding-bottom:0.4ex;
}

checkbutton {
	min-height:30px;
}

checkbutton check {
	min-height:17px;
	min-width:17px;
}

/* Make focus more visible (useful for keyboard navigation) */
radio, check, buttonbox > button, .dialogbutton, combobox button, spinbutton, switch, date_entry, time_entry, duration_entry {
	border-width:1px;
	padding:1px;
}

/* These rules for "hidden" elements fix GTK warnings about negative content
   width for Date/Time widgets that are contained in revealers */
.hidden {
	border-width:0;
	padding:0;
}

.hidden.dialog_error {
	transition-duration:0;
}

radio:focus, check:focus, buttonbox > button:focus, .dialogbutton:focus, combobox button:focus, spinbutton:focus, switch:focus, notebook:focus tab:checked, .focus {
	border-color:#398ee7;
	border-width:2px;
	padding:0px;
}

/* We removed padding from dialog button; to compensate add to inner label */
.dialogbutton > label {
	padding-right:0.6em;
	padding-left:0.6em;
}

/* Remove button transitions - can cause needless redraws */
buttonbox > button, .dialogbutton {
	transition-duration:0s;
	transition-delay:0s;
}

.textentry, .longtextentry {
	border-width:1px;
	padding:1px 5px;
}

.textentry {
	min-width:17em;
}

.textentry:focus, .longtextentry.focus {
	border-color:#398ee7;
	border-width:2px;
	padding:0 4px;
}

.longtextentry {
	min-width:21em;
	padding-right:1px;
	background:white;
	border-radius:3px;
	min-height:6.4em;
}

.longtextentry > textview {
	padding-top:.3ex;
	padding-bottom:.2ex;
}

.longtextentry.focus {
	padding-right:0;
}

/* Tabs need a bit more formatting */
tab {
	padding-left:11px;
	padding-right:11px;
}

/* This selects the tab with focus */
notebook:focus tab:checked {
	border-style:solid;
	padding-left:10px;
	padding-right:10px;
}

/* Set height of Repeats tab, so expanding it doesn't expand Event dialog */
#grid-repeats {
	min-height:30.2ex; /* Works for my laptop - might need tweaking */
}

.llabel {
	min-height:30px;
	padding-right:0.4ex;
	padding-left:0.4ex;
}

.tlabel {
	min-height:1.4em;
	padding-right:0.4ex;
	padding-left:0.4ex;
}

.plabel {
	min-height:1.4em;
	min-width:8em;
	padding-right:0.4ex;
	padding-left:0.4ex;
}

.sectlabel {
	padding-bottom:.9ex;
}

.alertlabel {
	font-style:italic;
	padding-top:.9ex;
	padding-bottom:.9ex;
}

/* This is to prevent visibility of transition when dialog is redisplayed */
entry {
	transition-duration:0s;
	transition-delay:0s;
}

.dialog_error {
	background-color:#e33;
	transition-duration:0.15s;
	transition-delay:0s;
}

/* Find results dialog */

.find_results {
	background:white;
	min-width:30em;
	min-height:10em;
}

.find_results label {
	padding:0.3ex;
}


/* List box - make row more highlighted when focused than when selected */
list row:selected {
	background:#7ab1e7
}

list row:focus {
	background:#398ee7
}
